import time

from throttled.storage.base import HitsWindow


class MemoryWindow(HitsWindow):
    __slots__ = ("__expire_at", "__hits")

    def __init__(self, duration: float):
        self.__expire_at = time.time() + duration
        self.__hits = 0
        super().__init__()

    def incr(self, hits: int = 1) -> int:
        self.__hits += hits
        return self.__hits

    def get_remaining_seconds(self) -> float:
        return self.__expire_at - time.time()



from throttled.models import Rate
from throttled.storage.memory.storage import DURATION_CALC, MemoryStorage
from throttled.strategy.base import StorageFactory, Strategy


class MemoryStorageFactory(StorageFactory):
    def get_storage_for_strategy(
        self, strategy: Strategy, limit: Rate
    ) -> MemoryStorage:
        try:
            return MemoryStorage(
                interval=limit.interval,
                duration_calc=DURATION_CALC[type(strategy)],
            )
        except KeyError as err:
            raise TypeError(
                f"There is not MemoryStorage for strategy {strategy.__class__} implemented yet!"
            ) from err



from abc import ABC, abstractmethod
from typing import Sequence, Union

from fastapi.exceptions import HTTPException
from starlette import status
from starlette.middleware.base import RequestResponseEndpoint
from starlette.requests import Request
from starlette.responses import Response

from throttled.exceptions import RateLimitExceeded
from throttled.limiter import Limiter
from throttled.strategy.base import Strategy


class HTTPLimitExceeded(HTTPException):
    def __init__(self, exc: RateLimitExceeded):
        self.retry_after = exc.retry_after
        if self.retry_after is None:
            headers = {}
        else:
            headers = {"Retry-After": str(self.retry_after)}
        super().__init__(
            status_code=status.HTTP_429_TOO_MANY_REQUESTS,
            detail=exc.detail,
            headers=headers,
        )


def response_from_exception(
    exc: Union[RateLimitExceeded, HTTPLimitExceeded]
) -> Response:
    return Response(
        status_code=status.HTTP_429_TOO_MANY_REQUESTS,
        content=exc.detail,
        headers={"Retry-After": str(exc.retry_after)},
    )


class FastAPILimiter:
    """First adapter between limiters APIs and FastAPI"""

    def __init__(self, strategy: Strategy):
        self.__limiter = Limiter(strategy)

    def limit(self, key: str):
        try:
            self.__limiter.limit(key)
        except RateLimitExceeded as exc:
            raise HTTPLimitExceeded(exc) from exc


class FastAPIRequestLimiter(ABC, FastAPILimiter):
    _ignored_paths: Sequence[str] = ("docs", "redoc", "favicon.ico", "openapi.json")

    def __init__(self, strategy: Strategy, response_factory=response_from_exception):
        super().__init__(strategy)
        self.__response_factory = response_factory

    def ignore_path(self, path: str):
        ignored = list(self._ignored_paths)
        ignored.append(path)
        self._ignored_paths = ignored

    async def dispatch(
        self, request: Request, call_next: RequestResponseEndpoint
    ) -> Response:
        try:
            path = str(request.url).replace(str(request.base_url), "")
            if path not in self._ignored_paths:
                self(request)
            return await call_next(request)
        except HTTPLimitExceeded as exc:
            return self.__response_factory(exc)

    @abstractmethod
    def __call__(self, request: Request):
        ...



from abc import ABC, abstractmethod

from throttled.exceptions import RateLimitExceeded
from throttled.models import Hit, Rate
from throttled.storage.base import Storage


class StorageFactory(ABC):
    @abstractmethod
    def get_storage_for_strategy(self, strategy: "Strategy", limit: Rate) -> Storage:
        ...


class Strategy:
    __slots__ = ("__limit", "__storage")

    def __init__(self, limit: Rate, storage_factory: StorageFactory):
        self.__limit = limit
        self.__storage = storage_factory.get_storage_for_strategy(self, limit)

    def maybe_block(self, hit: Hit):
        """
        :param hit: The hit to be tested
        :raises: RateLimitExceeded
        """
        window = self.__storage.get_current_window(hit)
        if window.incr(hit.cost) > self.__limit.hits:
            raise RateLimitExceeded(
                hit.key,
                retry_after=window.get_remaining_seconds(),
            )



from starlette.requests import Request

from ..base import FastAPIRequestLimiter


class TotalLimiter(FastAPIRequestLimiter):
    def __call__(self, request: Request):
        self.limit(key="total")



import asyncio
import functools
from typing import Any, Callable, TypeVar

from throttled.models import Hit
from throttled.strategy import Strategy

FuncT = TypeVar("FuncT", bound=Callable[..., Any])


class Limiter:
    def __init__(self, strategy: Strategy):
        self.__strategy = strategy

    def limit(self, key: str):
        self.__strategy.maybe_block(Hit(key=key))

    def decorate(self, func: FuncT) -> FuncT:
        key = f"func_name={func.__name__}:func_hash={hash(func)}"

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            self.limit(key)
            return func(*args, **kwargs)

        @functools.wraps(func)
        async def a_wrapper(*args, **kwargs):
            self.limit(key)
            return await func(*args, **kwargs)

        return a_wrapper if asyncio.iscoroutinefunction(func) else wrapper



from abc import ABC, abstractmethod

from throttled.models import Hit


class HitsWindow:
    @abstractmethod
    def get_remaining_seconds(self) -> int:
        ...

    @abstractmethod
    def incr(self, hits: int = 1) -> int:
        ...


class Storage(ABC):
    @abstractmethod
    def get_current_window(self, hit: Hit) -> HitsWindow:
        ...



from typing import Callable, Mapping, MutableMapping, Type

from throttled.models import Hit
from throttled.storage.base import Storage
from throttled.storage.memory.window import MemoryWindow
from throttled.strategy import FixedWindowStrategy, MovingWindowStrategy
from throttled.strategy.base import Strategy


def _fixed_window_duration_calc(hit_time: float, interval: float) -> float:
    return hit_time - hit_time % interval + interval


def _moving_window_duration_calc(_: float, interval: float) -> float:
    return interval


_DurationCalcType = Callable[[float, float], float]
DURATION_CALC: Mapping[Type[Strategy], _DurationCalcType] = {
    FixedWindowStrategy: _fixed_window_duration_calc,
    MovingWindowStrategy: _moving_window_duration_calc,
}


class MemoryStorage(Storage):
    __slots__ = ("__interval", "__cache", "__duration_calc")

    def __init__(self, interval: float, duration_calc: _DurationCalcType):
        self.__cache: MutableMapping[str, MemoryWindow] = {}
        self.__interval = interval
        self.__duration_calc = duration_calc

    def get_current_window(self, hit: Hit) -> MemoryWindow:
        window = self.__cache.get(hit.key)
        if window is None or window.get_remaining_seconds() < 0:
            self.__cache[hit.key] = window = MemoryWindow(
                duration=self.__duration_calc(hit.time, self.__interval)
            )
        return window



from .base import FastAPILimiter, FastAPIRequestLimiter
from .limiters.host import HostBasedLimiter
from .limiters.total import TotalLimiter
from .utils import split_dependencies_and_middlewares

__all__ = [
    "TotalLimiter",
    "HostBasedLimiter",
    "split_dependencies_and_middlewares",
    "FastAPILimiter",
    "FastAPIRequestLimiter",
]
